//-----------------------------------------------------------------------------
// Created on: 13 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

#ifndef Aux_SharedPtr_HeaderFile
#define Aux_SharedPtr_HeaderFile

// CCDist includes
#include <Aux_SharedObject.h>

//! Shared pointer wrapping shared objects.
template <class SharedType>
class Aux_SharedPtr
{
// Construction & destruction:
public:

  //! Default constructor.
  Aux_SharedPtr<SharedType>()
  {
    m_pRef = NULL;
  }

  //! Constructor accepting the raw reference to the target object to wrap with
  //! a smart pointer.
  //! \param ref [in] reference to the target object.
  Aux_SharedPtr<SharedType>(SharedType* ref)
  {
    m_pRef = ref;
    this->beginScope();
  }

  //! Constructor accepting the smart reference to the target object to share.
  //! \param smart_ref [in] smart reference to the target object.
  Aux_SharedPtr<SharedType>(const Aux_SharedPtr<SharedType>& smart_ref)
  {
    m_pRef = NULL;
    this->operator=(smart_ref);
  }

  //! Destructor.
  virtual ~Aux_SharedPtr()
  {
    this->endScope();
  }

public:

  //! Assignment operator accepting raw pointer.
  //! \param ref [in] raw pointer.
  //! \return this instance.
  Aux_SharedPtr<SharedType>& operator=(SharedType* ref)
  {
    this->endScope();
    m_pRef = ref;
    this->beginScope();
    return *this;
  }

  //! Assignment operator accepting smart pointer.
  //! \param smart_ref [in] smart pointer.
  //! \return this instance.
  Aux_SharedPtr<SharedType>& operator=(const Aux_SharedPtr<SharedType>& smart_ref)
  {
    this->endScope();
    m_pRef = smart_ref.m_pRef;
    this->beginScope();
    return *this;
  }

  //! Conversion operator.
  //! \return raw pointer.
  operator SharedType*()
  {
    return m_pRef;
  }

  //! Conversion operator.
  //! \return raw pointer.
  operator const SharedType*() const
  {
    return m_pRef;
  }

  //! Nullifies smart pointer.
  //! \return true/false.
  inline void Nullify()
  {
    this->endScope();
  }

  //! Checks whether the actually wrapper pointer is null or not.
  //! \return true/false.
  inline bool IsNull() const
  {
    return m_pRef == NULL;
  }

  //! Accessor for the wrapped pointer.
  //! \return wrapped pointer.
  inline SharedType* Access() const
  {
    return m_pRef;
  }

  //! Dereferencing operator.
  //! \return wrapped pointer.
  inline SharedType* operator->() const
  {
    return this->Access();
  }

public:

  //! Casts the passed pointer to the necessary type.
  //! \param smart_ref [in] pointer to cast.
  //! \return casted pointer.
  static Aux_SharedPtr<SharedType> DownCast(const Aux_SharedPtr<Aux_SharedObject>& smart_ref)
  {
    return Aux_SharedPtr<SharedType>( smart_ref.IsNull() ? NULL : dynamic_cast<SharedType*>( smart_ref.Access() ) );
  }

  //! Casts the passed pointer to the necessary type.
  //! \param smart_ref [in] pointer to cast.
  //! \return casted pointer.
  template <class OtherType>
  static Aux_SharedPtr<SharedType> DownCast(const Aux_SharedPtr<OtherType>& smart_ref)
  {
    return Aux_SharedPtr<SharedType>( smart_ref.IsNull() ? NULL : dynamic_cast<SharedType*>( smart_ref.Access() ) );
  }

private:

  //! Increments reference counter in the handled object.
  void beginScope()
  {
    if ( m_pRef )
      m_pRef->IncRef();
  }

  //! Decrements reference counter in the handled object.
  //! If counter's value goes to 0, the object is deleted.
  void endScope()
  {
    if ( m_pRef )
    {
      m_pRef->DecRef();
      if ( !m_pRef->NbRefs() )
      {
        delete m_pRef;
        m_pRef = NULL;
      }
    }
  }

private:

  SharedType* m_pRef; //!< Actual reference to the wrapped object.

};

//-----------------------------------------------------------------------------
// Handy short-cuts
//-----------------------------------------------------------------------------

//! Short-cut for shared pointer.
#define smptr Aux_SharedPtr

#endif
