//-----------------------------------------------------------------------------
// Created on: 13 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

// Own include
#include <BSpl_BasisEffectiveN.h>

// STD includes
#include <string.h>
#include <math.h>

//! Constructor.
BSpl_BasisEffectiveN::BSpl_BasisEffectiveN()
{}

//! Destructor.
BSpl_BasisEffectiveN::~BSpl_BasisEffectiveN()
{}

//! Evaluates non-vanishing basis functions.
//!
//! \param u  [in]  parameter value to evaluate basis functions in.
//! \param pU [in]  knot vector.
//! \param p  [in]  desired degree of basis functions to evaluate.
//! \param i  [in]  span index the parameter u falls to.
//! \param N  [out] evaluated functions. Note that the invoker code
//!                 must allocate memory for this array. Its dimension is (p + 1).
void BSpl_BasisEffectiveN::operator()(const double  u,
                                      const double* pU,
                                      const int     p,
                                      const int     i,
                                      double*       N) const
{
  // Initial basis function to start iterative evaluation
  N[0] = 1.0;

  // According to computational theory for effective B-spline evaluation,
  // we need to allocate p items for left[] and right[] terms. We will never
  // need left[0] and right[0] terms as these terms vanish due to the
  // organization of calculation scheme. Nevertheless, we prefer to allocate
  // memory for these item as well, just because it is more convenient and
  // more consistent with theory to start indexation from 1
  double* left = new double[p+1]; memset( left, 0, (p+1)*sizeof(double) );
  double* right = new double[p+1]; memset( right, 0, (p+1)*sizeof(double) );

  // We start from degree equal to 1. This is because basis functions
  // are constant and equal to 1 for degree = 0. We elevate degree re-using
  // previously calculated B-spline values according to the recurrent
  // formulation given in theory
  for ( int deg = 1; deg <= p; ++deg )
  {
    left[deg] = u - pU[i + 1 - deg];
    right[deg] = pU[i + deg] - u;

    // This variable contains value reused on adjacent iterations
    // by the number of evaluated functions
    double savedTerm = 0.0;

    // Now we iterate over the number of evaluated functions. Notice that
    // even though we have (deg + 1) of such functions, we iterate only deg
    // times as the last function can be simply evaluated thanks to savedTerm
    // variable without additional efforts (see theory)
    for ( int idx = 0; idx < deg; ++idx )
    {
      double rightCoeff = N[idx] / (right[idx+1] + left[deg-idx]);
      N[idx] = savedTerm + right[idx+1]*rightCoeff;
      savedTerm = left[deg-idx]*rightCoeff;
    }
    N[deg] = savedTerm;
  }

  // Release resources
  delete[] left;
  delete[] right;
}
