//-----------------------------------------------------------------------------
// Created on: 13 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

#ifndef BSpl_BasisEffectiveN_HeaderFile
#define BSpl_BasisEffectiveN_HeaderFile

//! Algorithm calculating those B-spline basis functions which are
//! non-vanishing in the given span. The target span is referenced by
//! its zero-based index. In order to use this functionality properly,
//! one should locate the target span by BasisFindSpan() and then ask this
//! routine to evaluate the basis functions for the passed parameter u.
//! This function performs effective evaluation, so as only those basis
//! functions which are not constant nulls are taken into account.
//! Moreover, this function reuses those values which are calculated at
//! previous stages of its internal iterations (the evaluation scheme
//! is principally sequential).
//!
//! The implemented approach is based on algorithm A2.2 from "The NURBS Book".
//! The main idea is to evaluate only those basis functions which give
//! any effect on the selected span. Such functions are all deduced from
//! the 0-degree basis function:
//!
//! <pre>
//!         N_i-p,p
//!          /
//!        ...
//!        /
//!     N_i-1,1
//!      /
//! N_i,0 --- span #i yields functions from (i-p) to p
//!      \
//!     N_i,1
//!        \
//!        ...
//!          \
//!         N_i,p
//! </pre>
//!
//! Indeed, for span {i} the only zero-degree non-vanishing basis function
//! is {N_i,0}. Now we want to know what are the non-vanishing functions
//! of degree {p}. From Cox-deBoor formulation we know that higher-degree
//! functions are expressed with lower-degree functions.
//!
//! According to the notes given in "The NURBS Book", it is possible
//! to re-use some terms during the evaluation procedure. The following
//! working variables are used:
//!
//! <pre>
//! left[j] = u - u_i+1-j
//! right[j] = u_i+j - u
//! </pre>
//!
//! Here i is the 0-based index of the working span.
class BSpl_BasisEffectiveN
{
// Construction & destruction:
public:

  BSpl_BasisEffectiveN();
  virtual ~BSpl_BasisEffectiveN();

public:

  void operator()(const double  u,
                  const double* pU,
                  const int     p,
                  const int     i,
                  double*       N) const;

};

#endif
