//-----------------------------------------------------------------------------
// Created on: 13 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

// Own include
#include <BSpl_BasisEffectiveNDers.h>

// Aux includes
#include <Aux_HeapAlloc.h>

//! Constructor.
BSpl_BasisEffectiveNDers::BSpl_BasisEffectiveNDers()
{}

//! Destructor.
BSpl_BasisEffectiveNDers::~BSpl_BasisEffectiveNDers()
{}

//! Evaluates non-vanishing basis functions with their derivatives of
//! (k)-th order.
//!
//! \param u      [in]  parameter value to evaluate functions in.
//! \param pU     [in]  knot vector.
//! \param p      [in]  desired degree of basis functions to evaluate.
//! \param span_i [in]  span index the parameter u falls to.
//! \param n      [in]  maximal order to derivatives to calculate.
//! \param ders   [out] evaluated functions. Note that the invoker code
//!                     must allocate memory for this array. Its dimension is
//!                     (p + 1)*(k + 1).
void BSpl_BasisEffectiveNDers::operator()(const double  u,
                                          const double* pU,
                                          const int     p,
                                          const int     span_i,
                                          const int     n,
                                          double**      ders) const
{
  Aux_HeapAlloc Alloc;
  Aux_HeapAlloc2D Alloc2D;

  // Array to store working matrix
  double** ndu = Alloc2D.Allocate<double>(p+1, p+1, true);

  //------------------------------------------------------
  // Evaluate basis functions along with knot differences
  //------------------------------------------------------

  // Initial basis function to start iterative evaluation
  ndu[0][0] = 1.0;

  // Arrays left[] and right[] are used to store knot differences
  double* left = Alloc.Allocate<double>(p+1, true);
  double* right = Alloc.Allocate<double>(p+1, true);

  // We start from degree equal to 1. This is because basis functions
  // are constant and equal to 1 for degree = 0. We elevate degree re-using
  // previously calculated B-spline values according to the recurrent
  // formulation given in theory
  for ( int deg = 1; deg <= p; ++deg )
  {
    left[deg] = u - pU[span_i + 1 - deg];
    right[deg] = pU[span_i + deg] - u;

    // This variable contains value reused on adjacent iterations
    // by the number of evaluated functions
    double savedTerm = 0.0;

    // Now we iterate over the number of evaluated functions. Notice that
    // even though we have (deg + 1) of such functions, we iterate only deg
    // times as the last function can be simply evaluated thanks to savedTerm
    // variable without additional efforts (see theory)
    for ( int idx = 0; idx < deg; ++idx )
    {
      // Knot differences (lower triangle)
      ndu[deg][idx] = right[idx+1] + left[deg-idx];
      double temp = ndu[idx][deg-1] / ndu[deg][idx];

      // Basis functions (upper triangle)
      ndu[idx][deg] = savedTerm + right[idx+1]*temp;
      savedTerm = left[deg-idx]*temp;
    }
    ndu[deg][deg] = savedTerm;
  }

  //----------------------
  // Evaluate derivatives
  //----------------------

  // Prepare matrix of coefficients
  double** a = Alloc2D.Allocate<double>(2, n+1, true); // Two working rows

  // Load basis functions into the first row of result matrix
  for ( int j = 0; j <= p; ++j )
    ders[0][j] = ndu[j][p]; // Take all effective functions of degree p

  // Loop over the effective indices
  for ( int r = 0; r <= p; ++r )
  {
    int s1 = 0, s2 = 1; // Alternate rows in array a
    a[0][0] = 1.0;

    // Loop to compute (k)-th derivative
    for ( int k = 1; k <= n; ++k )
    {
      double d = 0.0; // Derivative being evaluated
      int rk = r-k, pk = p-k;

      if ( r >= k )
      {
        a[s2][0] = a[s1][0]/ndu[pk+1][rk];
        d = a[s2][0]*ndu[rk][pk];
      }

      int j1, j2;
      if ( rk >= -1 )
        j1 = 1;
      else
        j1 = -rk;
      if ( r-1 <= pk )
        j2 = k-1;
      else
        j2 = p-r;

      for ( int j = j1; j <= j2; ++j )
      {
        a[s2][j] = (a[s1][j]-a[s1][j-1])/ndu[pk+1][rk+j];
        d += a[s2][j]*ndu[rk+j][pk];
      }

      if ( r <= pk )
      {
        a[s2][k] = -a[s1][k-1]/ndu[pk+1][r];
        d += a[s2][k]*ndu[r][pk];
      }

      ders[k][r] = d;
      int j = s1; s1 = s2; s2 = j; // Switch rows
    }
  }

  // Multiply through by the correct factors
  int r = p;
  for ( int k = 1; k <= n; ++k )
  {
    for ( int j = 0; j <= p; ++j )
      ders[k][j] *= r;

    r *= (p-k);
  }
}
