//-----------------------------------------------------------------------------
// Created on: 09 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

// Own include
#include <Extrema_CurveCurve.h>

// Extrema includes
#include <Extrema_DistWithPenalty.h>

// Opt includes
#include <Opt_GradientDescent.h>
#include <Opt_PSO.h>

// STD includes
#include <iostream>

//-----------------------------------------------------------------------------
// Target function
//-----------------------------------------------------------------------------

//! Distance function with gradient.
class Extrema_DistFunc_Grad : public Opt_FunctionWithGradient<uv>
{
public:

  Extrema_DistFunc_Grad(const smptr<bcurve>& C1,
                        const smptr<bcurve>& C2) : m_C1(C1), m_C2(C2)
  {}

public:

  inline virtual double Value(const uv& x)
  {
    double u;
    if ( x.U() < m_C1->MinParameter() )
      u = m_C1->MinParameter();
    else if ( x.U() > m_C1->MaxParameter() )
      u = m_C1->MaxParameter();
    else
      u = x.U();

    double v;
    if ( x.V() < m_C2->MinParameter() )
      v = m_C2->MinParameter();
    else if ( x.V() > m_C2->MaxParameter() )
      v = m_C2->MaxParameter();
    else
      v = x.V();

    xyz P1, P2;
    m_C1->Eval(u, P1);
    m_C2->Eval(v, P2);

    return (P1 - P2).Dot(P1 - P2);
  }

  inline virtual t_coord Gradient(const uv& x)
  {
    double u;
    if ( x.U() < m_C1->MinParameter() )
      u = m_C1->MinParameter();
    else if ( x.U() > m_C1->MaxParameter() )
      u = m_C1->MaxParameter();
    else
      u = x.U();

    double v;
    if ( x.V() < m_C2->MinParameter() )
      v = m_C2->MinParameter();
    else if ( x.V() > m_C2->MaxParameter() )
      v = m_C2->MaxParameter();
    else
      v = x.V();

    xyz P1, P2;
    m_C1->Eval(u, P1);
    m_C2->Eval(v, P2);

    xyz G1, G2;
    m_C1->Eval_Dk(u, 1, G1);
    m_C2->Eval_Dk(v, 1, G2);

    const double grad_u =  (P1 - P2).Dot(G1)*2;
    const double grad_v = -(P1 - P2).Dot(G2)*2;

    return uv(grad_u, grad_v);
  }

protected:

  smptr<bcurve> m_C1, m_C2;

};

//-----------------------------------------------------------------------------
// Curve-curve extrema
//-----------------------------------------------------------------------------

//! Creates penalty function at the given point.
//! \param p         [in] point to create a penalty for.
//! \param radius    [in] radius of penalty splash.
//! \param amplifier [in] amplification factor.
smptr<Opt_Splash_NM>
  Extrema_CurveCurve::Penalty(const uv&    p,
                              const double radius,
                              const double amplifier)
{
  const double u_peak = p.U();
  const double v_peak = p.V();

  smptr<Opt_Splash_N> N;
  {
    const double U[] = {0.0, 0.0, 0.0, 0.0,
                        u_peak - radius,
                        u_peak - 0.5*radius,
                        u_peak,
                        u_peak + 0.5*radius,
                        u_peak + radius,
                        1.0, 1.0, 1.0, 1.0};
    const int nU = sizeof(U) / sizeof(double);
    const int p = 3;
    N = new Opt_Splash_N(U, nU, p, 4);
  }

  smptr<Opt_Splash_N> M;
  {
    const double U[] = {0.0, 0.0, 0.0, 0.0,
                        v_peak - radius,
                        v_peak - 0.5*radius,
                        v_peak,
                        v_peak + 0.5*radius,
                        v_peak + radius,
                        1.0, 1.0, 1.0, 1.0};
    const int nU = sizeof(U) / sizeof(double);
    const int p = 3;
    M = new Opt_Splash_N(U, nU, p, 4);
  }

  smptr<Opt_Splash_NM> P = new Opt_Splash_NM(N, M, amplifier);
  return P;
}

//! Default constructor.
Extrema_CurveCurve::Extrema_CurveCurve()
{}

//! Constructor performing full initialization.
//! \param C1 [in] first curve.
//! \param C2 [in] second curve.
Extrema_CurveCurve::Extrema_CurveCurve(const smptr<bcurve>& C1,
                                       const smptr<bcurve>& C2)
{
  this->Init(C1, C2);
}

//! Initializes extrema tool with operands.
//! \param C1 [in] first curve.
//! \param C2 [in] second curve.
void Extrema_CurveCurve::Init(const smptr<bcurve>& C1,
                              const smptr<bcurve>& C2)
{
  m_C1 = C1;
  m_C2 = C2;
}

//! Searches for the nearest points.
//! \return true in case of success, false -- otherwise.
bool Extrema_CurveCurve::Perform()
{
  //-------------------------------------------------------------------------
  // PSO optimization
  //-------------------------------------------------------------------------

  Extrema_DistWithPenalty F_PSO(m_C1, m_C2);

  // Smoothness
  const Prim_Cx cont_C1  = m_C1->Continuity();
  const Prim_Cx cont_C2  = m_C2->Continuity();
  const bool    isSmooth = (cont_C1*cont_C2 > 0);

  if ( !isSmooth )
  {
    // TODO: cout is a bad practice (some logger to be used instead)
    std::cout << "The method is not suitable for C0 geometry" << std::endl;
    return false;
  }

  // PSO parameters
  Opt_PSO<uv>::t_search_params pso_params;
  pso_params.num_particles   = 300;      // In PSO the number of agents is typically small
  pso_params.num_iterations  = 5000;     // More than enough in typical situations
  pso_params.area.corner_min = uv(0, 0); // Min corner of the search area
  pso_params.area.corner_max = uv(1, 1); // Max corner of the search area
  pso_params.precision       = 1.0e-6;   // Less values are of no practical sense
  pso_params.pFunc           = &F_PSO;   // Objective function
  pso_params.m               = 0.729;    // Retain weight
  pso_params.mu_cognition    = 0.0;      // Pure social behavior (no "nostalgia")
  pso_params.nu_social       = 1.49445;  // Social determinant of an agent

  // Parameter of the scheme
  const double amplifier    = 10.0; // Depends on the problem
  const double radius       = 0.2;  // Depends on the problem
  const double dist_barrier = 10.0; // Depends on the problem
  const int    max_attempts = 10;   // Limit max number of extrema to find

  // Working variable
  double prev_dist = 0.0;
  int    attempt   = 0;
  uv     prev_sol;

  // Loop on global search
  do
  {
    ++attempt;

    // Run PSO
    int pso_num_iters = 0;
    Opt_PSO<uv> PSO(pso_params);
    PSO.Perform(pso_num_iters);
    const Opt_PSO<uv>::t_measuring& pso_sol = PSO.GetBestGlobal();

    // TODO: cout is a bad practice (some logger to be used instead)
    //std::cout << "Consumed number of iterations: " << pso_num_iters                          << std::endl;
    //std::cout << "Distance: "                      << pso_sol.Q                              << std::endl;
    //std::cout << "Parameters: "                    << pso_sol.p.U() << ", " << pso_sol.p.V() << std::endl;

    // Add penalty splash
    F_PSO.AddPenalty( Penalty(pso_sol.p, radius, amplifier) );

    //-------------------------------------------------------------------------
    // Gradient descent
    //-------------------------------------------------------------------------

    double u_sol, v_sol;
    Extrema_DistFunc_Grad F_Grad(m_C1, m_C2);

    // Gradient descent parameters
    Opt_GradientDescent<uv>::t_search_params grad_params;
    grad_params.num_iterations = 10000;
    grad_params.pFunc          = &F_Grad;
    grad_params.precision      = 1.0e-9;
    grad_params.start          = pso_sol.p;
    grad_params.is_adaptive    = true;
    grad_params.step           = 1.0e-1; // To be corrected by Armijo rule

    // Set coarse solution
    m_coarseSolutions.push_back(pso_sol.p);

    // Run local optimization
    int grad_num_iters = 0;
    Opt_GradientDescent<uv> Descent(grad_params);
    if ( !Descent.Perform(grad_num_iters) )
    {
      // TODO: cout is a bad practice (some logger to be used instead)
      std::cout << "Warning: gradient descent did not converge" << std::endl;
      // Still, let's continue for better robustness of the method
    }
    const uv& grad_sol = Descent.Solution();

    // TODO: cout is a bad practice (some logger to be used instead)
    //std::cout << "Consumed number of iterations (gradient descent): " << grad_num_iters << std::endl;

    // Store solution
    u_sol = grad_sol.U();
    v_sol = grad_sol.V();

    // Fit ranges
    uv normalized_sol = this->fit(u_sol, v_sol);

    // TODO: cout is a bad practice (some logger to be used instead)
    //std::cout << "Parameters: " << u_sol << ", " << v_sol << std::endl;
    //std::cout << "---"                                    << std::endl;

    // Measure distance
    xyz P1, P2;
    m_C1->Eval(normalized_sol.U(), P1);
    m_C2->Eval(normalized_sol.V(), P2);
    const double dist = (P1 - P2).Modulus();

    if ( attempt > 1 && fabs(dist - prev_dist) > dist_barrier )
      break; // Stop when other solutions start to be inappropriate

    // Add solution
    if ( !this->addSolution(normalized_sol, pso_params.precision*10) )
      break;

    prev_dist = dist;
  }
  while ( attempt < max_attempts );

  return m_solutions.size() > 0;
}

//! Fits the passed values of (U, V) into their corresponding parametric ranges.
//! \param u [in] U value.
//! \param v [in] V value.
//! \return normalized solution.
uv Extrema_CurveCurve::fit(const double u,
                           const double v) const
{
  double nu = u, nv = v;

  if ( u < m_C1->MinParameter() )
    nu = m_C1->MinParameter();
  else if ( u > m_C1->MaxParameter() )
    nu = m_C1->MaxParameter();

  if ( v < m_C2->MinParameter() )
    nv = m_C2->MinParameter();
  else if ( v > m_C2->MaxParameter() )
    nv = m_C2->MaxParameter();

  return uv(nu, nv);
}

//! Checks whether the passed solution is to be added (duplications are
//! avoided). If not, returns false, otherwise returns true.
//! \param sol       [in] solution to check and add.
//! \param precision [in] precision for equality check.
//! \return true/false.
bool Extrema_CurveCurve::addSolution(const uv&    sol,
                                     const double precision)
{
  bool isDuplicate = false;
  for ( size_t s = 0; s < m_solutions.size(); ++s )
  {
    if ( (m_solutions[s] - sol).SquaredModulus() < precision )
    {
      isDuplicate = true;
      break;
    }
  }

  if ( !isDuplicate )
    m_solutions.push_back(sol);

  return !isDuplicate;
}
