//-----------------------------------------------------------------------------
// Created on: 09 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

#ifndef Extrema_CurveCurve_HeaderFile
#define Extrema_CurveCurve_HeaderFile

// Opt includes
#include <Opt_BSplineSplash.h>

// Prim includes
#include <Prim_BSplineCurve.h>
#include <Prim_UV.h>

//! This tool searches for the nearest points between two curves using
//! PSO scheme described in the article "Parallel Reverse Niching PSO for
//! Multimodal Optimization" by E. Zafarani-Moattar, H. Haj-Seyed-Javadi.
//! The following modifications are also adopted:
//!
//! - The process has two stages. PSO is used to find coarse solutions which
//!   are then precised using local search (simplest way is to run gradient
//!   descent).
//!
//! - B-spline product is used as a penalty function. Comparing to Gauss
//!   function used in the original articles, B-spline product offers better
//!   control over the niching form (however, it is less efficient in terms
//!   of computational costs).
//!
//! The weak point of the method is the adopted niching technique which in
//! most cases performs one redundant optimization session. This drawback
//! is natural to the applied scheme as it is hard to know when to stop
//! putting penalties. As we do not know this, we have to run optimization
//! until we are starting to get the same solutions. Such a stopping criterion
//! is greedy, so it is better to use some heuristics to limit the number
//! of optimization attempts. E.g. if you search for exactly one solution,
//! you'd better to limit the number of attempts with 1.
//!
//! Another possible solution is to avoid using penalties and start using
//! sub-swarms. The latter, however, involves sub-swarming techniques which
//! may appear to be costly (e.g. those based on computation of distances
//! between the agents on each iteration/epoch). This is to be checked.
class Extrema_CurveCurve
{
public:

  static smptr<Opt_Splash_NM>
    Penalty(const uv&    p,
            const double radius,
            const double amplifier);

public:

  Extrema_CurveCurve();
  Extrema_CurveCurve(const smptr<bcurve>& C1,
                     const smptr<bcurve>& C2);

public:

  void Init(const smptr<bcurve>& C1,
            const smptr<bcurve>& C2);

  bool Perform();

public:

  //! \return solutions.
  inline const std::vector<uv>& Solutions() const
  {
    return m_solutions;
  }

  //! \return coarse solutions.
  inline const std::vector<uv>& CoarseSolutions() const
  {
    return m_coarseSolutions;
  }

protected:

  uv fit(const double u,
         const double v) const;

  bool addSolution(const uv&    sol,
                   const double precision);

private:

  smptr<bcurve>   m_C1;              //!< First curve to analyze.
  smptr<bcurve>   m_C2;              //!< Second curve to analyze.
  std::vector<uv> m_solutions;       //!< Solutions.
  std::vector<uv> m_coarseSolutions; //!< Coarse solutions.

};

#endif
