//-----------------------------------------------------------------------------
// Created on: 13 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

// Windows includes
#include <windows.h>

// Aux includes
#include <Aux_String.h>

// Extrema includes
#include <Extrema_CurveCurve.h>

// STD includes
#include <iostream>
#include <fstream>

#define BUFSIZE     1000
#define CCDIST_DATA "CCDIST_DATA"

//! Entry point.
int main(int argc, char** argv)
{
  /* =======================
   *  Load curves from file
   * ======================= */

  if ( argc != 2 )
  {
    std::cout << "Error: invalid number of arguments";
    std::cout << " (filename is expected)" << std::endl;
    return 1;
  }

  std::vector< smptr<bcurve> > curves;

  TCHAR chNewEnv[BUFSIZE];
  GetEnvironmentVariable(CCDIST_DATA, chNewEnv, BUFSIZE);

  // Open file with input data
  std::string filename = Aux::slashed( std::string(chNewEnv) ) + argv[1];
  std::ifstream FILE(filename);
  if ( !FILE.is_open() )
    return false;

  while ( !FILE.eof() )
  {
    char str[256];
    FILE.getline(str, 256);

    std::vector<std::string> tokens;
    std::istringstream iss(str);
    std::copy( std::istream_iterator<std::string>(iss),
               std::istream_iterator<std::string>(),
               std::back_inserter< std::vector<std::string> >(tokens) );

    if ( tokens.empty() )
      continue;

    int deg = 0;
    std::vector<double> U;
    std::vector<xyz> Poles;

    enum e_mode
    {
      mode_undefined = 0,
      mode_degree,
      mode_knots,
      mode_poles
    } mode = mode_undefined;

    for ( int idx = 1; idx < (int) tokens.size(); ++idx )
    {
      // Degree
      if ( tokens[idx] == "-degree" )
      {
        mode = mode_degree;
        continue;
      }
      if ( mode == mode_degree && Aux::is_number( tokens[idx] ) )
      {
        deg = Aux::to_number<int>(tokens[idx]);
        continue;
      }

      // Knots
      if ( tokens[idx] == "-knots" )
      {
        mode = mode_knots;
        continue;
      }
      if ( mode == mode_knots )
      {
        U.push_back( Aux::to_number<double>(tokens[idx]) );
        continue;
      }

      // Poles
      if ( tokens[idx] == "-poles" )
      {
        mode = mode_poles;
        continue;
      }
      if ( mode == mode_poles )
      {
        double x = Aux::to_number<double>(tokens[idx]);
        double y = Aux::to_number<double>(tokens[++idx]);
        double z = Aux::to_number<double>(tokens[++idx]);

        Poles.push_back( xyz(x, y, z) );
        continue;
      }
    }

    // ...
    // Check control equation m = n + deg + 1
    // ...

    const int m = (int) U.size() - 1;
    const int n_plus_p_plus_1 = (int) Poles.size() + deg;

    if ( m != n_plus_p_plus_1 )
    {
      std::cout << "Error: m = n + deg + 1 is not conformed" << std::endl;
      std::cout << "\tm is " << m << std::endl;
      std::cout << "\tn is " << Poles.size() << std::endl;
      std::cout << "\tn + deg + 1 is " << n_plus_p_plus_1 << std::endl;
      return 1;
    }

    // Construct B-spline curve
    smptr<bcurve> crv = new bcurve(Poles, U, deg);
    curves.push_back(crv);
  }

  // Close file
  FILE.close();

  if ( curves.size() != 2 )
  {
    std::cout << "Error: expected number of curves is 2" << std::endl;
    return 1;
  }

  /* ==========================
   *  Calculate extrema points
   * ========================== */

  Extrema_CurveCurve ExtCC(curves[0], curves[1]);
  if ( !ExtCC.Perform() )
  {
    std::cout << "Error: cannot calculate extrema" << std::endl;
    return 1;
  }
  else
  {
    const std::vector<uv>& solutions = ExtCC.Solutions();
    for ( size_t k = 0; k < solutions.size(); ++k )
    {
      const double sol_u = solutions[k].U();
      const double sol_v = solutions[k].V();

      xyz P1, P2;
      curves[0]->Eval(sol_u, P1);
      curves[1]->Eval(sol_v, P2);

      std::cout << "\nSolution #" << (k + 1) << ":" << std::endl;
      std::cout << "Parameter on C1: " << sol_u << std::endl;
      std::cout << "Parameter on C2: " << sol_v << std::endl;
      std::cout << "Point on C1: (" << P1.X() << ", " << P1.Y() << ", " << P1.Z() << ")" << std::endl;
      std::cout << "Point on C2: (" << P2.X() << ", " << P2.Y() << ", " << P2.Z() << ")" << std::endl;
      std::cout << "Distance: " << (P1 - P2).Modulus() << std::endl;
    }
  }

  return 0;
}
