//-----------------------------------------------------------------------------
// Created on: 08 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

#ifndef Opt_ArmijoRule_HeaderFile
#define Opt_ArmijoRule_HeaderFile

// Opt includes
#include <Opt_FunctionWithGradient.h>

//! Armijo rule for adaptive selection of step length in inexact line search.
template <typename T_DIMENSION>
class Opt_ArmijoRule
{
public:

  typedef T_DIMENSION t_coord;

public:

  //! Search parameters.
  struct t_search_params
  {
    double                             max_alpha;      //!< Right bound.
    int                                num_iterations; //!< Max number of iterations.
    t_coord                            x_k;            //!< Position in the search space.
    t_coord                            d_k;            //!< Direction of line search.
    Opt_FunctionWithGradient<t_coord>* pFunc;          //!< Target function.

    t_search_params() : max_alpha      (0.0),
                        num_iterations (0),
                        pFunc          (NULL)
    {}
  };

public:

  Opt_ArmijoRule() {}

public:

  //! Runs Armijo rule.
  //! \param params    [in]  search parameters.
  //! \param num_iters [out] consumed number of iterations.
  //! \param alpha     [out] resulting step size.
  //! \return true in case of success, false -- otherwise.
  bool Perform(const t_search_params& params,
               int&                   num_iters,
               double&                alpha)
  {
    num_iters = 0;
    alpha     = params.max_alpha;

    const double  beta      = 0.2; // Interval reduction coefficient
    const double  mu        = 1.0e-2;
    const double  min_alpha = 1.0e-8;
    const t_coord gradient  = params.pFunc->Gradient(params.x_k);
    const double  f         = params.pFunc->Value(params.x_k);
    const double  phi_deriv = gradient.Dot(params.d_k);

    // Initial alpha
    bool isSolved = false;
    bool doStop   = false;

    // Main iterations
    do
    {
      ++num_iters;
      if ( num_iters < params.num_iterations )
      {
        const double barrier = alpha*mu*phi_deriv;
        const double phi     = params.pFunc->Value(params.x_k + params.d_k*alpha) - f;

        if ( phi < barrier )
        {
          doStop   = true;
          isSolved = true;
        }
        else
        {
          alpha *= beta;

          if ( alpha < min_alpha )
          {
            doStop   = true;
            isSolved = false;
          }
        }
      }
      else
      {
        doStop   = true;
        isSolved = false;
      }
    }
    while ( !doStop );

    return isSolved;
  }

};

#endif
