//-----------------------------------------------------------------------------
// Created on: 12 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

#ifndef Opt_BSplineSplash_HeaderFile
#define Opt_BSplineSplash_HeaderFile

// Aux includes
#include <Aux_SharedObject.h>

// Opt includes
#include <Opt_Function.h>

// Prim includes
#include <Prim_UV.h>

// BSpl includes
#include <BSpl_BasisN.h>

//! Basis one-dimensional spline.
class Opt_Splash_N : public Aux_SharedObject
{
public:

  //! Constructor accepting B-spline parameters.
  //! \param pU  [in] knots.
  //! \param nU  [in] number of knots.
  //! \param p   [in] B-spline degree.
  //! \param idx [in] 0-based index of a B-spline function in series.
  Opt_Splash_N(const double* pU,
               const int     nU,
               const int     p,
               const int     idx)
  {
    m_iU   = nU;
    m_iDeg = p;
    m_iIdx = idx;

    // Copy knot vector
    m_pU = new double[nU];
    memcpy_s( m_pU, nU*sizeof(double), pU, nU*sizeof(double) );
  }

  //! Destructor.
  ~Opt_Splash_N()
  {
    delete[] m_pU;
  }

public:

  //! Evaluates function for the given value.
  //! \param u [in] value to evaluate the function at.
  //! \return evaluated value.
  double Value(const double u)
  {
    return m_eval(u, m_pU, m_iU, m_iDeg, m_iIdx);
  }

private:

  double*     m_pU;   //!< Knots.
  int         m_iU;   //!< Number of knots.
  int         m_iDeg; //!< Degree.
  int         m_iIdx; //!< Index of the B-spline function.
  BSpl_BasisN m_eval; //!< B-spline evaluation tool.
};

//! Product of two basis splines.
class Opt_Splash_NM : public Opt_Function<uv>
{
public:

  //! Constructor accepting two basis splines to build a product from.
  //! \param pN        [in] first basis function.
  //! \param pM        [in] second basis function.
  //! \param amplifier [in] amplification coefficient.
  Opt_Splash_NM(const smptr<Opt_Splash_N>& pN,
                const smptr<Opt_Splash_N>& pM,
                const double               amplifier)
  : m_pN   (pN),
    m_pM   (pM),
    m_fAmp (amplifier)
  {}

public:

  //! Evaluates splash function.
  //! \param p [in] point to evaluate the function in.
  //! \return function value.
  virtual double Value(const uv& p)
  {
    const double n = m_pN->Value( p.U() );
    const double m = m_pM->Value( p.V() );
    return n*m*m_fAmp;
  }

private:

  smptr<Opt_Splash_N> m_pN;   //!< First basis spline in a product.
  smptr<Opt_Splash_N> m_pM;   //!< Second basis spline in a product.
  double              m_fAmp; //!< Amplification factor.

};

#endif
