//-----------------------------------------------------------------------------
// Created on: 07 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

#ifndef Opt_GradientDescent_HeaderFile
#define Opt_GradientDescent_HeaderFile

// Opt includes
#include <Opt_ArmijoRule.h>

//! Gradient descent optimization.
template <typename T_DIMENSION>
class Opt_GradientDescent
{
public:

  typedef T_DIMENSION t_coord;

public:

  //! Search parameters.
  struct t_search_params
  {
    t_coord                            start;          //!< Starting point.
    int                                num_iterations; //!< Max number of iterations.
    double                             precision;      //!< Precision stopping criterion.
    double                             step;           //!< Step size.
    bool                               is_adaptive;    //!< Whether to use adaptive step.
    Opt_FunctionWithGradient<t_coord>* pFunc;          //!< Target function.

    t_search_params() : num_iterations (0),
                        precision      (0.0),
                        step           (0.0),
                        is_adaptive    (false),
                        pFunc          (NULL)
    {}
  };

public:

  //! Constructor accepting search parameters.
  //! \param params [in] search parameters.
  Opt_GradientDescent(const t_search_params& params) : m_params(params) {}

public:

  //! \return solution.
  inline const t_coord& Solution() const
  {
    return m_argmin;
  }

  //! \return sampling points.
  inline const std::vector<t_coord>& Samples() const
  {
    return m_samples;
  }

public:

  //! Runs gradient descent optimization.
  //! \param num_iters [out] consumed number of iterations.
  //! \return true in case of success, false -- otherwise.
  bool Perform(int& num_iters)
  {
    num_iters     = 0;
    t_coord pos   = m_params.start;
    bool doStop   = false;
    bool isSolved = false;

    // Prepare Armijo rule
    Opt_ArmijoRule<t_coord> Armijo;
    Opt_ArmijoRule<t_coord>::t_search_params Armijo_params;
    Armijo_params.max_alpha      = m_params.step;
    Armijo_params.num_iterations = 100;
    Armijo_params.pFunc          = m_params.pFunc;

    // Main iterations
    do
    {
      ++num_iters;
      if ( num_iters < m_params.num_iterations )
      {
        m_samples.push_back(pos);

        t_coord g = m_params.pFunc->Gradient(pos);
        doStop = ( g.Modulus() < m_params.precision );

        if ( !doStop )
        {
          double actual_step = 0.0;

          // Use Armijo rule
          if ( m_params.is_adaptive )
          {
            Armijo_params.x_k = pos;
            Armijo_params.d_k = g.Invert(); // Anti-gradient is the search direction
            int num_armijo_iters = 0;

            if ( !Armijo.Perform(Armijo_params, num_armijo_iters, actual_step) )
              actual_step = m_params.step;
          }
          else
            actual_step = m_params.step;

          // Update rule
          pos = pos - g*actual_step;

          //std::cout << "F[" << num_iters << ", step " << actual_step
          //          << "] = " << m_params.pFunc->Value(pos) << std::endl;
        }
        else
        {
          m_argmin = pos;
          isSolved = true;
        }
      }
      else
      {
        m_argmin = pos;
        doStop   = true;
      }
    }
    while ( !doStop );

    return isSolved;
  }

private:

  std::vector<t_coord> m_samples; //!< Sampling points.
  t_search_params      m_params;  //!< Search parameters.
  t_coord              m_argmin;  //!< Solution.

};

#endif
