//-----------------------------------------------------------------------------
// Created on: 13 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

// Own include
#include <Prim_BSplineCurve.h>

// Aux includes
#include <Aux_HeapAlloc.h>

// BSpl includes
#include <BSpl_BasisEffectiveN.h>
#include <BSpl_BasisEffectiveNDers.h>
#include <BSpl_BasisFindSpan.h>

// STD includes
#include <algorithm>

//! Constructor.
//! \param Poles [in] poles for B-spline curve.
//! \param pU    [in] knot vector.
//! \param nU    [in] number of knots.
//! \param p     [in] degree.
Prim_BSplineCurve::Prim_BSplineCurve(const std::vector<Prim_XYZ>& Poles,
                                     const double*                pU,
                                     const int                    nU,
                                     const int                    p)
: Prim_Curve(), m_pU(NULL)
{
  this->init(Poles, pU, nU, p);
}

//! Constructor.
//! \param Poles [in] poles for B-spline curve.
//! \param U     [in] knot vector.
//! \param p     [in] degree.
Prim_BSplineCurve::Prim_BSplineCurve(const std::vector<Prim_XYZ>& Poles,
                                     const std::vector<double>&   U,
                                     const int                    p)
: Prim_Curve(), m_pU(NULL)
{
  const int nU = (int) U.size();
  double* U_raw = new double[nU];
  for ( int i = 0; i < nU; ++i )
    U_raw[i] = U[i];

  this->init(Poles, U_raw, nU, p);

  delete[] U_raw;
}

//! Destructor.
Prim_BSplineCurve::~Prim_BSplineCurve()
{
  delete[] m_pU; m_pU = NULL;
}

//! Returns first knot.
//! \return first knot.
double Prim_BSplineCurve::MinParameter() const
{
  return m_pU[0];
}

//! Returns last knot.
//! \return last knot.
double Prim_BSplineCurve::MaxParameter() const
{
  return m_pU[m_iU-1];
}

//! Evaluates B-spline curve for the given parameter.
//! \param u [in]  parameter value to evaluate curve for.
//! \param P [out] 3D point corresponding to the given parameter on curve.
void Prim_BSplineCurve::Eval(const double u,
                             Prim_XYZ&    P) const
{
  // Heap allocator
  Aux_HeapAlloc Alloc;

  // Find span the passed u falls into
  BSpl_BasisFindSpan FindSpan;
  int span = FindSpan(u, m_pU, m_iU, m_iDeg);

  // Evaluate effective B-spline Basis functions
  BSpl_BasisEffectiveN EffectiveN;
  double* N = Alloc.Allocate<double>(m_iDeg + 1, true);
  EffectiveN(u, m_pU, m_iDeg, span, N);

  // Evaluate curve
  Prim_XYZ C;
  for ( int i = 0; i <= m_iDeg; ++i )
  {
    C += m_poles[span-m_iDeg+i].Multiplied(N[i]); // See theory for clarifications on indices
  }

  // Set output parameter
  P = C;
}

//! Evaluates derivative of the B-spline curve for the given parameter.
//! \param u   [in]  parameter value to evaluate curve for.
//! \param k   [in]  the desired order of derivative.
//! \param dkC [out] derivative vector.
void Prim_BSplineCurve::Eval_Dk(const double u,
                                const int    k,
                                Prim_XYZ&    dkC) const
{
  Aux_HeapAlloc2D Alloc;
  double** dN = Alloc.Allocate<double>(m_iDeg + 1, m_iDeg + 1, true);
  Eval_Dk(dN, u, k, dkC);
}

//! Evaluates derivative of the B-spline curve for the given parameter.
//! \param dN  [in]  array for results (can be passed for better performance).
//! \param u   [in]  parameter value to evaluate curve for.
//! \param k   [in]  the desired order of derivative.
//! \param d1C [out] derivative vector.
void Prim_BSplineCurve::Eval_Dk(double**     dN,
                                const double u,
                                const int    k,
                                Prim_XYZ&    d1C) const
{
  // Find span the passed u falls into
  BSpl_BasisFindSpan FindSpan;
  int span = FindSpan(u, m_pU, m_iU, m_iDeg);

  // Evaluate derivatives of B-spline basis functions
  BSpl_BasisEffectiveNDers NDers;
  NDers(u, m_pU, m_iDeg, span, m_iDeg, dN);

  // Evaluate curve
  Prim_XYZ C;
  for ( int i = 0; i <= m_iDeg; ++i )
  {
    C += m_poles[span-m_iDeg+i].Multiplied(dN[k][i]); // See theory for clarifications on indices
  }

  // Set output parameter
  d1C = C;
}

//! Returns continuity of the curve.
//! \return continuity.
Prim_Cx Prim_BSplineCurve::Continuity() const
{
  std::vector<int> mults;
  int mult = 1;
  for ( int i = 0; i < m_iU; ++i )
  {
    if ( m_pU[i] == this->MinParameter() )
      continue;

    if ( fabs( m_pU[i] - m_pU[i - 1] ) < DBL_EPSILON )
      ++mult;
    else if ( m_pU[i - 1] != this->MinParameter() )
    {
      mults.push_back(mult);
      mult = 1;

      if ( m_pU[i] == this->MaxParameter() )
        break;
    }
  }

  if ( mults.size() )
  {
    const int max_mult = *std::max_element( mults.begin(), mults.end() );
    const int cont = m_iDeg - max_mult;

    if ( cont <= 0 )
      return Smoothness_C0;
    if ( cont == 1 )
      return Smoothness_C1;
    if ( cont == 2 )
      return Smoothness_C2;
    if ( cont == 3 )
      return Smoothness_C3;
  }

  return Smoothness_CN;
}

//! Initializes B-spline curve with complete data.
//! \param Poles [in] control points.
//! \param pU    [in] knot vector.
//! \param nU    [in] number of knots.
//! \param p     [in] degree of the B-spline basis functions.
void Prim_BSplineCurve::init(const std::vector<Prim_XYZ>& Poles,
                             const double*                pU,
                             const int                    nU,
                             const int                    p)
{
  m_poles = Poles;
  m_iU    = nU;
  m_iDeg  = p;

  // Copy knot vector
  delete[] m_pU;
  m_pU = new double[nU];
  memcpy_s( m_pU, nU*sizeof(double), pU, nU*sizeof(double) );
}
