//-----------------------------------------------------------------------------
// Created on: 13 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

// Own include
#include <Prim_XYZ.h>

// STD includes
#include <math.h>

//-----------------------------------------------------------------------------
// Class-level API
//-----------------------------------------------------------------------------

//! Returns (0, 0, 0).
//! \return origin.
Prim_XYZ Prim_XYZ::O()
{
  return Prim_XYZ();
}

//! Returns (1, 0, 0).
//! \return direction for OX axis.
Prim_XYZ Prim_XYZ::OX()
{
  return Prim_XYZ(1.0, 0.0, 0.0);
}

//! Returns (0, 1, 0).
//! \return direction for OY axis.
Prim_XYZ Prim_XYZ::OY()
{
  return Prim_XYZ(0.0, 1.0, 0.0);
}

//! Returns (0, 0, 1).
//! \return direction for OZ axis.
Prim_XYZ Prim_XYZ::OZ()
{
  return Prim_XYZ(0.0, 0.0, 1.0);
}

//-----------------------------------------------------------------------------
// Object-level API
//-----------------------------------------------------------------------------

//! Default constructor. Initializes point co-ordinates with origin values:
//! (0, 0, 0).
Prim_XYZ::Prim_XYZ()
{
  m_fX = m_fY = m_fZ = 0.0;
}

//! Constructor with coordinates.
//! \param x [in] first co-ordinate.
//! \param y [in] second co-ordinate.
//! \param z [in] third co-ordinate.
Prim_XYZ::Prim_XYZ(const double x,
                       const double y,
                       const double z)
{
  m_fX = x;
  m_fY = y;
  m_fZ = z;
}

//! Assignment constructor.
//! \param XYZ [in] point to assign to this one.
Prim_XYZ::Prim_XYZ(const Prim_XYZ& XYZ)
{
  this->operator=(XYZ);
}

//! Destructor.
Prim_XYZ::~Prim_XYZ()
{
}

//! Returns modulus of the point's radius vector.
//! \return modulus.
double Prim_XYZ::Modulus() const
{
  return sqrt( this->SquaredModulus() );
}

//! Returns squared modulus of the point's radius vector.
//! \return modulus.
double Prim_XYZ::SquaredModulus() const
{
  return m_fX*m_fX + m_fY*m_fY + m_fZ*m_fZ;
}

//! Multiplies copy of point by the passed scalar value.
//! \param coeff [in] scalar value to multiply point by.
//! \return resulting point.
Prim_XYZ Prim_XYZ::Multiplied(const double coeff) const
{
  return this->operator*(coeff);
}

//! Normalizes point so that it gets unit length.
void Prim_XYZ::Normalize()
{
  const double modulus = this->Modulus();
  m_fX /= modulus;
  m_fY /= modulus;
  m_fZ /= modulus;
}

//! Creates a normalized copy of this point.
//! \return normalized copy.
Prim_XYZ Prim_XYZ::Normalized() const
{
  Prim_XYZ C(m_fX, m_fY, m_fZ);
  C.Normalize();
  return C;
}

//! Calculates dot product between this and another vector.
//! \param XYZ [in] another vector.
//! \return dot product.
double Prim_XYZ::Dot(const Prim_XYZ& XYZ) const
{
  return m_fX*XYZ.m_fX + m_fY*XYZ.m_fY + m_fZ*XYZ.m_fZ;
}

//! Calculates cross product between this and another vector.
//! \param XYZ [in] another vector.
//! \return cross product.
Prim_XYZ Prim_XYZ::Cross(const Prim_XYZ& XYZ) const
{
  const double x = m_fY*XYZ.m_fZ - m_fZ*XYZ.m_fY;
  const double y = m_fZ*XYZ.m_fX - m_fX*XYZ.m_fZ;
  const double z = m_fX*XYZ.m_fY - m_fY*XYZ.m_fX;

  return Prim_XYZ(x, y, z);
}

//! Calculates angle between two points.
//! \param XYZ [in] another vector.
//! \return angle in radians.
double Prim_XYZ::Angle(const Prim_XYZ& XYZ) const
{
  Prim_XYZ V1 = this->Normalized();
  Prim_XYZ V2 = XYZ.Normalized();

  double cos_alpha = V1.Dot(V2);

  // Compensate rounding errors
  if ( cos_alpha > 1 )
    cos_alpha = 1.0;

  // Compensate rounding errors
  if ( cos_alpha < -1 )
    cos_alpha = -1.0;

  const double alpha = acos(cos_alpha);
  return alpha;
}

//! Assignment operator.
//! \param XYZ [in] point to copy into this one.
//! \return this one.
Prim_XYZ& Prim_XYZ::operator=(const Prim_XYZ& XYZ)
{
  m_fX = XYZ.m_fX;
  m_fY = XYZ.m_fY;
  m_fZ = XYZ.m_fZ;
  return *this;
}

//! Multiplies copy of point by the passed scalar value.
//! \param coeff [in] scalar value to multiply point by.
//! \return resulting point.
Prim_XYZ Prim_XYZ::operator*(const double coeff) const
{
  Prim_XYZ XYZ_Copy(*this);
  XYZ_Copy *= coeff;
  return XYZ_Copy;
}

//! Multiplies this point by the passed scalar value.
//! \param coeff [in] scalar value to multiply point by.
//! \return this point multiplied by the passed scalar.
Prim_XYZ Prim_XYZ::operator*=(const double coeff)
{
  this->m_fX *= coeff;
  this->m_fY *= coeff;
  this->m_fZ *= coeff;
  return *this;
}

//! Divides copy of point by the passed scalar value.
//! \param coeff [in] scalar value to divide point by.
//! \return resulting point.
Prim_XYZ Prim_XYZ::operator/(const double coeff) const
{
  Prim_XYZ XYZ_Copy(*this);
  XYZ_Copy /= coeff;
  return XYZ_Copy;
}

//! Divides this point by the passed scalar value.
//! \param coeff [in] scalar value to divide point by.
//! \return this point multiplied by the passed scalar.
Prim_XYZ Prim_XYZ::operator/=(const double coeff)
{
  this->m_fX /= coeff;
  this->m_fY /= coeff;
  this->m_fZ /= coeff;
  return *this;
}

//! Adds the passed point to the copy of this one.
//! \param XYZ [in] point to add.
//! \return result of addition.
Prim_XYZ Prim_XYZ::operator+(const Prim_XYZ& XYZ) const
{
  Prim_XYZ XYZ_Copy(*this);
  XYZ_Copy += XYZ;
  return XYZ_Copy;
}

//! Adds the passed point to this one.
//! \param XYZ [in] point to add.
//! \return result of addition.
Prim_XYZ& Prim_XYZ::operator+=(const Prim_XYZ& XYZ)
{
  this->m_fX += XYZ.m_fX;
  this->m_fY += XYZ.m_fY;
  this->m_fZ += XYZ.m_fZ;
  return *this;
}

//! Inverts the copy of this point.
//! \return result of inversion.
Prim_XYZ Prim_XYZ::Invert() const
{
  Prim_XYZ XYZ_Copy(*this);
  XYZ_Copy.m_fX = -this->m_fX;
  XYZ_Copy.m_fY = -this->m_fY;
  XYZ_Copy.m_fZ = -this->m_fZ;
  return XYZ_Copy;
}

//! Subtracts the passed point from the copy of this one.
//! \param XYZ [in] point to subtract.
//! \return result of subtraction.
Prim_XYZ Prim_XYZ::operator-(const Prim_XYZ& XYZ) const
{
  return this->operator+( XYZ.Invert() );
}

//! Adds the passed point to this one.
//! \param XYZ [in] point to add.
//! \return result of subtraction.
Prim_XYZ& Prim_XYZ::operator-=(const Prim_XYZ& XYZ)
{
  return this->operator+=( XYZ.Invert() );
}
