//-----------------------------------------------------------------------------
// Created on: 13 October 2015
// Created by: Sergey SLYADNEV
// Web-site: http://quaoar.su/blog
//-----------------------------------------------------------------------------

#ifndef Prim_XYZ_HeaderFile
#define Prim_XYZ_HeaderFile

// STD includes
#include <float.h>
#include <math.h>
#include <stddef.h>

//! Geometric primitive for 3D point.
class Prim_XYZ
{
// Static API:
public:

  static int num_coordinates()
  {
    return 3;
  }

  static Prim_XYZ O();
  static Prim_XYZ OX();
  static Prim_XYZ OY();
  static Prim_XYZ OZ();

// Construction & destruction:
public:

  Prim_XYZ();

  Prim_XYZ(const double x,
           const double y,
           const double z);

  Prim_XYZ(const Prim_XYZ& XYZ);

  virtual ~Prim_XYZ();

public:

  //! Returns X coordinate of the 3D point.
  //! \return X coordinate.
  inline double X() const
  {
    return m_fX;
  }

  //! Sets X coordinate.
  //! \param x [in] value to set.
  inline void SetX(const double x)
  {
    m_fX = x;
  }

  //! Returns Y coordinate of the 3D point.
  //! \return Y coordinate.
  inline double Y() const
  {
    return m_fY;
  }

  //! Sets Y coordinate.
  //! \param y [in] value to set.
  inline void SetY(const double y)
  {
    m_fY = y;
  }

  //! Returns Z coordinate of the 3D point.
  //! \return Z coordinate.
  inline double Z() const
  {
    return m_fZ;
  }

  //! Sets Z coordinate.
  //! \param z [in] value to set.
  inline void SetZ(const double z)
  {
    m_fZ = z;
  }

  //! Returns coordinate by its 0-based index.
  //! \param idx [in] 0 for X, 1 for Y, 2 for Z.
  //! \return requested coordinate.
  inline double Coord(const int idx) const
  {
    if ( idx == 0 )
      return this->X();

    if ( idx == 1 )
      return this->Y();

    if ( idx == 2 )
      return this->Z();

    return DBL_MAX;
  }

  //! Updates coordinate having the specified 0-based index with the
  //! passed value.
  //! \param idx [in] 0 for X, 1 for Y, 2 for Z.
  //! \param val [in] value to set.
  inline double SetCoord(const int idx,
                         const double val)
  {
    double* coord = NULL;

    if ( idx == 0 )
      coord = &m_fX;
    if ( idx == 1 )
      coord = &m_fY;
    if ( idx == 2 )
      coord = &m_fZ;

    if ( coord )
      *coord = val;

    return DBL_MAX;
  }

  //! Returns true if this vector is zero.
  //! \param tol3D [in] three-dimensional tolerance for comparison.
  //! \return true/false.
  inline bool IsOrigin(const double tol3D = 0.0) const
  {
    return ( fabs(m_fX) <= tol3D ) && ( fabs(m_fY) <= tol3D ) && ( fabs(m_fZ) <= tol3D );
  }

public:

  double    Modulus        ()                    const;
  double    SquaredModulus ()                    const;
  Prim_XYZ  Multiplied     (const double coeff)  const;
  void      Normalize      ();
  Prim_XYZ  Normalized     ()                    const;
  double    Dot            (const Prim_XYZ& XYZ) const;
  Prim_XYZ  Cross          (const Prim_XYZ& XYZ) const;
  double    Angle          (const Prim_XYZ& XYZ) const;

public:

  Prim_XYZ& operator=  (const Prim_XYZ& XYZ);
  Prim_XYZ  operator*  (const double coeff)   const;
  Prim_XYZ  operator*= (const double coeff);
  Prim_XYZ  operator/  (const double coeff)   const;
  Prim_XYZ  operator/= (const double coeff);
  Prim_XYZ  operator+  (const Prim_XYZ& XYZ)  const;
  Prim_XYZ& operator+= (const Prim_XYZ& XYZ);
  Prim_XYZ  Invert()                          const;
  Prim_XYZ  operator-  (const Prim_XYZ& XYZ)  const;
  Prim_XYZ& operator-= (const Prim_XYZ& XYZ);

private:

  double m_fX; //!< X coordinate.
  double m_fY; //!< Y coordinate.
  double m_fZ; //!< Z coordinate.

};

//! Convenience alias.
typedef Prim_XYZ xyz;

#endif
